use crate::geometry::Scale;

/// A 1-dimensional scale.
pub type Scale1<T> = Scale<T, 1>;

/// A 2-dimensional scale.
pub type Scale2<T> = Scale<T, 2>;

/// A 3-dimensional scale.
pub type Scale3<T> = Scale<T, 3>;

/// A 4-dimensional scale.
pub type Scale4<T> = Scale<T, 4>;

/// A 5-dimensional scale.
pub type Scale5<T> = Scale<T, 5>;

/// A 6-dimensional scale.
pub type Scale6<T> = Scale<T, 6>;
