// Generated by gir (https://github.com/gtk-rs/gir @ c93f8f31b6a5)
// from  (@ 3879ff8e0201)
// from gir-files (@ 01ea5f1a07cd)
// DO NOT EDIT

#include "manual.h"
#include <stdio.h>

#define PRINT_CONSTANT(CONSTANT_NAME) \
    printf("%s;", #CONSTANT_NAME); \
    printf(_Generic((CONSTANT_NAME), \
                    char *: "%s", \
                    const char *: "%s", \
                    char: "%c", \
                    signed char: "%hhd", \
                    unsigned char: "%hhu", \
                    short int: "%hd", \
                    unsigned short int: "%hu", \
                    int: "%d", \
                    unsigned int: "%u", \
                    long: "%ld", \
                    unsigned long: "%lu", \
                    long long: "%lld", \
                    unsigned long long: "%llu", \
                    float: "%.6f", \
                    double: "%.6f", \
                    long double: "%ld"), \
           CONSTANT_NAME); \
    printf("\n");

int main() {
    PRINT_CONSTANT((gint) GLY_LOADER_ERROR_FAILED);
    PRINT_CONSTANT((gint) GLY_LOADER_ERROR_NO_MORE_FRAMES);
    PRINT_CONSTANT((gint) GLY_LOADER_ERROR_UNKNOWN_IMAGE_FORMAT);
    PRINT_CONSTANT((gint) GLY_MEMORY_A8B8G8R8);
    PRINT_CONSTANT((gint) GLY_MEMORY_A8R8G8B8);
    PRINT_CONSTANT((gint) GLY_MEMORY_A8R8G8B8_PREMULTIPLIED);
    PRINT_CONSTANT((gint) GLY_MEMORY_B8G8R8);
    PRINT_CONSTANT((gint) GLY_MEMORY_B8G8R8A8);
    PRINT_CONSTANT((gint) GLY_MEMORY_B8G8R8A8_PREMULTIPLIED);
    PRINT_CONSTANT((gint) GLY_MEMORY_G16);
    PRINT_CONSTANT((gint) GLY_MEMORY_G16A16);
    PRINT_CONSTANT((gint) GLY_MEMORY_G16A16_PREMULTIPLIED);
    PRINT_CONSTANT((gint) GLY_MEMORY_G8);
    PRINT_CONSTANT((gint) GLY_MEMORY_G8A8);
    PRINT_CONSTANT((gint) GLY_MEMORY_G8A8_PREMULTIPLIED);
    PRINT_CONSTANT((gint) GLY_MEMORY_R16G16B16);
    PRINT_CONSTANT((gint) GLY_MEMORY_R16G16B16A16);
    PRINT_CONSTANT((gint) GLY_MEMORY_R16G16B16A16_FLOAT);
    PRINT_CONSTANT((gint) GLY_MEMORY_R16G16B16A16_PREMULTIPLIED);
    PRINT_CONSTANT((gint) GLY_MEMORY_R16G16B16_FLOAT);
    PRINT_CONSTANT((gint) GLY_MEMORY_R32G32B32A32_FLOAT);
    PRINT_CONSTANT((gint) GLY_MEMORY_R32G32B32A32_FLOAT_PREMULTIPLIED);
    PRINT_CONSTANT((gint) GLY_MEMORY_R32G32B32_FLOAT);
    PRINT_CONSTANT((gint) GLY_MEMORY_R8G8B8);
    PRINT_CONSTANT((gint) GLY_MEMORY_R8G8B8A8);
    PRINT_CONSTANT((gint) GLY_MEMORY_R8G8B8A8_PREMULTIPLIED);
    PRINT_CONSTANT((guint) GLY_MEMORY_SELECTION_A8B8G8R8);
    PRINT_CONSTANT((guint) GLY_MEMORY_SELECTION_A8R8G8B8);
    PRINT_CONSTANT((guint) GLY_MEMORY_SELECTION_A8R8G8B8_PREMULTIPLIED);
    PRINT_CONSTANT((guint) GLY_MEMORY_SELECTION_B8G8R8);
    PRINT_CONSTANT((guint) GLY_MEMORY_SELECTION_B8G8R8A8);
    PRINT_CONSTANT((guint) GLY_MEMORY_SELECTION_B8G8R8A8_PREMULTIPLIED);
    PRINT_CONSTANT((guint) GLY_MEMORY_SELECTION_G16);
    PRINT_CONSTANT((guint) GLY_MEMORY_SELECTION_G16A16);
    PRINT_CONSTANT((guint) GLY_MEMORY_SELECTION_G16A16_PREMULTIPLIED);
    PRINT_CONSTANT((guint) GLY_MEMORY_SELECTION_G8);
    PRINT_CONSTANT((guint) GLY_MEMORY_SELECTION_G8A8);
    PRINT_CONSTANT((guint) GLY_MEMORY_SELECTION_G8A8_PREMULTIPLIED);
    PRINT_CONSTANT((guint) GLY_MEMORY_SELECTION_R16G16B16);
    PRINT_CONSTANT((guint) GLY_MEMORY_SELECTION_R16G16B16A16);
    PRINT_CONSTANT((guint) GLY_MEMORY_SELECTION_R16G16B16A16_FLOAT);
    PRINT_CONSTANT((guint) GLY_MEMORY_SELECTION_R16G16B16A16_PREMULTIPLIED);
    PRINT_CONSTANT((guint) GLY_MEMORY_SELECTION_R16G16B16_FLOAT);
    PRINT_CONSTANT((guint) GLY_MEMORY_SELECTION_R32G32B32A32_FLOAT);
    PRINT_CONSTANT((guint) GLY_MEMORY_SELECTION_R32G32B32A32_FLOAT_PREMULTIPLIED);
    PRINT_CONSTANT((guint) GLY_MEMORY_SELECTION_R32G32B32_FLOAT);
    PRINT_CONSTANT((guint) GLY_MEMORY_SELECTION_R8G8B8);
    PRINT_CONSTANT((guint) GLY_MEMORY_SELECTION_R8G8B8A8);
    PRINT_CONSTANT((guint) GLY_MEMORY_SELECTION_R8G8B8A8_PREMULTIPLIED);
    PRINT_CONSTANT((gint) GLY_SANDBOX_SELECTOR_AUTO);
    PRINT_CONSTANT((gint) GLY_SANDBOX_SELECTOR_BWRAP);
    PRINT_CONSTANT((gint) GLY_SANDBOX_SELECTOR_FLATPAK_SPAWN);
    PRINT_CONSTANT((gint) GLY_SANDBOX_SELECTOR_NOT_SANDBOXED);
    return 0;
}
