// This file was generated by gir (https://github.com/gtk-rs/gir)
// from gir-files (https://github.com/gtk-rs/gir-files)
// DO NOT EDIT

mod desktop_app_info;
pub use self::desktop_app_info::DesktopAppInfo;

mod fd_message;
pub use self::fd_message::FDMessage;

mod file_descriptor_based;
pub use self::file_descriptor_based::FileDescriptorBased;

mod input_stream;
pub use self::input_stream::InputStream;

mod mount_monitor;
pub use self::mount_monitor::MountMonitor;

mod output_stream;
pub use self::output_stream::OutputStream;

#[cfg(feature = "v2_84")]
#[cfg_attr(docsrs, doc(cfg(feature = "v2_84")))]
mod mount_entry;
#[cfg(feature = "v2_84")]
#[cfg_attr(docsrs, doc(cfg(feature = "v2_84")))]
pub use self::mount_entry::MountEntry;

mod mount_point;
pub use self::mount_point::MountPoint;

pub(crate) mod functions;

pub(crate) mod traits {
    pub use super::fd_message::FDMessageExt;
    pub use super::file_descriptor_based::FileDescriptorBasedExt;
    pub use super::input_stream::UnixInputStreamExt;
    pub use super::output_stream::UnixOutputStreamExt;
}
