// This file was generated by gir (https://github.com/gtk-rs/gir)
// from
// from gir-files
// DO NOT EDIT

use glib::translate::*;

use crate::ffi;

glib::wrapper! {
    pub struct Cicp(BoxedInline<ffi::GlyCicp>);

    match fn {
        copy => |ptr| ffi::gly_cicp_copy(mut_override(ptr)),
        free => |ptr| ffi::gly_cicp_free(ptr),
        type_ => || ffi::gly_cicp_get_type(),
    }
}

unsafe impl Send for Cicp {}
unsafe impl Sync for Cicp {}
