// This file was generated by gir (https://github.com/gtk-rs/gir)
// from gir-files (https://github.com/gtk-rs/gir-files)
// DO NOT EDIT

use crate::{FileDescriptorBased, ffi};
use glib::{prelude::*, translate::*};

glib::wrapper! {
    #[doc(alias = "GUnixInputStream")]
    pub struct InputStream(Object<ffi::GUnixInputStream, ffi::GUnixInputStreamClass>) @extends gio::InputStream, @implements gio::PollableInputStream, FileDescriptorBased;

    match fn {
        type_ => || ffi::g_unix_input_stream_get_type(),
    }
}

impl InputStream {
    pub const NONE: Option<&'static InputStream> = None;
}

pub trait UnixInputStreamExt: IsA<InputStream> + 'static {
    #[doc(alias = "g_unix_input_stream_get_close_fd")]
    #[doc(alias = "get_close_fd")]
    #[doc(alias = "close-fd")]
    fn closes_fd(&self) -> bool {
        unsafe {
            from_glib(ffi::g_unix_input_stream_get_close_fd(
                self.as_ref().to_glib_none().0,
            ))
        }
    }
}

impl<O: IsA<InputStream>> UnixInputStreamExt for O {}
